import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import seaborn as sns
import xgboost as xgb
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import accuracy_score, classification_report, balanced_accuracy_score, roc_auc_score, precision_recall_curve, auc, confusion_matrix
import json
from sklearn.model_selection import GridSearchCV
from sklearn.preprocessing import StandardScaler
import shap
from tqdm import tqdm

def shorten_cfa_name(name):
    if len(name) > 30:
        # Trouver le premier espace au milieu du texte
        words = name.split()
        mid = len(words) // 2
        return " ".join(words[:mid]) + "\n" + " ".join(words[mid:])
    return name

class XGBModel:
    def __init__(self, X, y, objective, eval_metric, test_size=0.15, random_state=42):

        self.X = X
        self.y = y
        self.test_size = test_size
        self.random_state = random_state

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            X, y, test_size=self.test_size, random_state=self.random_state, stratify=y
        )

        neg, pos = np.bincount(self.y_train)
        self.class_ratio = neg / pos

        self.objective = objective
        self.eval_metric = eval_metric

    def corr_plot(self) : 

        corr_matrix = self.X.corr()
        # On garde uniquement la moitié inférieure pour éviter les doublons
        mask = np.tril(np.ones(corr_matrix.shape), k=-1).astype(bool)
        corr_pairs = corr_matrix.where(mask)

        # On sélectionne uniquement les corrélations égales à 1 (ou très proches de 1 à cause de la précision numérique)
        correlated_vars = corr_pairs.stack().reset_index()
        correlated_vars.columns = ['Variable_1', 'Variable_2', 'Correlation']

        # On filtre les corrélations parfaites
        perfect_corr = correlated_vars[correlated_vars['Correlation'] == 1.0]

        print(perfect_corr)

        fig , ax = plt.subplots(figsize = (30,20))
        corr_matrix = self.X.corr()
        sns.heatmap(corr_matrix, annot=False)
        plt.show()

    def grid_search(self, objective = "binary:logistic", eval_metric = 'logloss', cv = 3) :

        param_grid = {
            "n_estimators": [100, 500],
            "learning_rate": [0.1, 0.05],
            "max_depth": [3, 5, 7],
            "subsample": [0.8, 1.0],
            "colsample_bytree": [0.8, 1.0],
            "reg_lambda": [0.5, 1],
            "reg_alpha": [0, 0.5]
        }

        grid_search = GridSearchCV(
            estimator=xgb.XGBClassifier(objective= objective, eval_metric= eval_metric),
            param_grid=param_grid,
            scoring="accuracy",
            cv=cv,  # Validation croisée à 3 folds
            n_jobs=-1,
            verbose=1
        )

        grid_search.fit(self.X_train, self.y_train)

        self.best_params_ = grid_search.best_params_
        self.best_model_ = grid_search.best_estimator_

        return self.best_params_
    
    def fit(self) : 

        if not hasattr(self, "best_params_"):
            raise ValueError("grid_search() must be called before fit().")
    
        else : 
            self.model = xgb.XGBClassifier(
                objective=self.objective,
                eval_metric=self.eval_metric,
                n_estimators= self.best_params_['n_estimators'],
                learning_rate=self.best_params_['learning_rate'],
                max_depth=self.best_params_['max_depth'], 
                colsample_bytree = self.best_params_['colsample_bytree'], 
                reg_alpha = self.best_params_['reg_alpha'] , 
                reg_lambda = self.best_params_['reg_lambda'],
                subsample = self.best_params_['subsample'], 
                scale_pos_weight= self.class_ratio)

            return self.model.fit(self.X_train, self.y_train)
        
    def predict(self, X):
        return self.model.predict(X)

    def predict_proba(self, X):
        return self.model.predict_proba(X)

    
    def scoring(self) : 

        # Prédictions sur le jeu de test (probabilités pour ROC-AUC et PR-AUC)
        self.y_pred = self.model.predict(self.X_test)
        self.y_pred_proba = self.model.predict_proba(self.X_test)[:, 1]  # Probabilité d'être classe 1

        # Accuracy classique
        accuracy = accuracy_score(self.y_test, self.y_pred)
        balanced_acc = balanced_accuracy_score(self.y_test, self.y_pred)

        # ROC-AUC score
        roc_auc = roc_auc_score(self.y_test, self.y_pred_proba)

        # PR-AUC (courbe Precision-Recall)
        precision, recall, _ = precision_recall_curve(self.y_test, self.y_pred_proba)
        pr_auc = auc(recall, precision)

        # Matrice de confusion
        conf_matrix = confusion_matrix(self.y_test, self.y_pred)

        self.accuracy = accuracy
        self.balanced_acc = balanced_acc
        self.roc_auc = roc_auc
        self.pr_auc = pr_auc
        self.classification_report = classification_report(self.y_test, self.y_pred)
        self.conf_matrix = conf_matrix

        # Affichage des résultats
        print(f"Accuracy: {accuracy:.2f}")
        print(f"Balanced Accuracy: {balanced_acc:.2f}")
        print(f"ROC-AUC Score: {roc_auc:.2f}")
        print(f"PR-AUC Score: {pr_auc:.2f}\n")

        print("Classification Report:")
        print(self.classification_report)

        print("Confusion Matrix:")
        print(conf_matrix)
    
    # ajouter les fonctions pour les visualisations

    def plot_importance(self) :
        
        fig, ax = plt.subplots(figsize = (8,10))
        xgb.plot_importance(self.model, ax = ax, height=0.9, grid=False, color = '#FFBF66', max_num_features=20)
        plt.show()
    
    def shap_values(self) :
        explainer = shap.TreeExplainer(self.model, approximate=True)
        self.shap_values = explainer(self.X_train)

        return self.shap_values

    def get_corr_shap_df(self) :
        corr_list = []
        feature_list = self.X_train.columns
        df_shap_values = pd.DataFrame(self.shap_values.values,
                                index = self.X_train.index,
                                columns = self.X_train.columns)

        # get correlation
        for feature in tqdm(feature_list) :
            b = np.corrcoef(df_shap_values[feature],self.X_train[feature])[1][0]
            corr_list.append(b)

        # correlation by feature
        df_corr = pd.concat([pd.Series(feature_list),pd.Series(corr_list)], axis = 1).fillna(0)
        df_corr.columns = ['Variable','Corr']

        # positive correlation in red and negative correlation in blue
        df_corr['Color_sign'] = np.where(df_corr['Corr'] > 0,'green','red')
        df_corr.loc[df_corr.Corr == 0,'Color_sign'] = 'grey'

        # plot it
        shap_abs = np.abs(df_shap_values)
        df_shap_abs = pd.DataFrame(shap_abs.mean()).reset_index()
        df_shap_abs.columns = ['Variable','SHAP_abs']

        # mean shap + corr
        df_corr_shap = pd.merge(df_shap_abs,df_corr, on = 'Variable')
        df_corr_shap['Sign_all'] = np.sign(df_corr_shap['Corr']) # get mean sign of correlation for total
        df_corr_shap = df_corr_shap.sort_values(by = 'SHAP_abs', ascending = False)

        self.df_plot = df_corr_shap.iloc[0:12,:].sort_values(by = 'SHAP_abs', ascending = True)
        self.df_plot['Variable'] = self.df_plot['Variable'].str.replace('_',' ')

        self.df_plot['SHAP_abs_sign'] = self.df_plot['SHAP_abs'] * self.df_plot['Sign_all']

        return self.df_plot

    def plot_impact(self) : 
        #color_map = {'Positif' : '#DE2E4B', 'Négatif' :'#82CEF9'}

        xlabel = 'Coefficient d\'impact'

        colors = []

        for value in self.df_plot['Corr'] : 
            if value > 0 : 
                colors.append('green')
            elif value < 0 : 
                colors.append('red')

        fig, ax = plt.subplots(figsize=(10,8))

        bars = ax.barh(self.df_plot['Variable'].apply(shorten_cfa_name), self.df_plot['SHAP_abs_sign'], color=colors, height=0.9)
        ax.set_xlabel(xlabel)
        ax.tick_params(axis='y', labelsize=7)

        #reste_index = self.df_plot[self.df_plot['index'] == f'{len(self.df_plot) - 20} autres variables'].index.values[0]
        #bars[reste_index].set_hatch('//')

        for n,bar in enumerate(bars) : 
            width = round(bar.get_width(),2)
            text = str(width)

            if colors[n] == 'green' : 
                ax.text(width*1.01, bar.get_y() + bar.get_height() / 2, text ,va='center', ha='left', color='green', fontsize=9)
            else : 
                ax.text(width*1.01, bar.get_y() + bar.get_height() / 2, text ,va='center', ha='right', color='red', fontsize=9)

        legend_elements = [
                mpatches.Patch(facecolor='green', label='Positif'),
                mpatches.Patch(facecolor='red', label='Négatif'),
            ]
        
        ax.vlines(0,-1, self.df_plot.shape[0], color = 'black')

        ax.set_xlim(min(self.df_plot['SHAP_abs_sign'])*1.2, max(self.df_plot['SHAP_abs_sign'])*1.3)
        ax.legend(handles=legend_elements, loc='best', title='Impact sur la prédiction')
        plt.title('Impact moyen des variables sur la prédiction (SHAP)', pad = 20, fontsize = 14)
        plt.tight_layout()
        plt.show()
        



            